"""
contains the MigXmlReader - a MIG Reader that is based on XML MIGs (and therefore requires lxml)
"""
import re
from pathlib import Path
from typing import List, Set, TypeVar, Union
from xml.etree.ElementTree import Element

try:
    from lxml import etree  # type:ignore[import]
except ImportError as import_error:
    import_error.msg += "; Did you install maus[xml]?"
    # lxml is only an optional dependency of maus but in this module, it is required
    raise

from more_itertools import first, one

from maus.edifact import EdifactFormat
from maus.models.edifact_components import EdifactStack, EdifactStackLevel
from maus.models.message_implementation_guide import SegmentGroupHierarchy
from maus.navigation import AhbLocation, AhbLocationLayer
from maus.reader.etree_element_helpers import get_nested_qualifiers
from maus.reader.mig_ahb_name_helpers import make_tree_names_comparable
from maus.reader.mig_reader import MigReader

Result = TypeVar("Result")  #: is a type var to indicate an "arbitrary but same" type in a generic function


def check_file_can_be_parsed_as_mig_xml(file_path: Path) -> None:
    """
    Returns nothing iff the given file is parsable as XML and contains no obvious errors.
    This is not a really sophisticated analysis but just a basic minimal sanity check.
    In case of error an exception is raised.
    """
    reader = MigXmlReader(file_path)
    _ = EdifactFormat(reader.get_format_name())  # dies with an exception if the value is invalid


# pylint:disable=c-extension-no-member
class MigXmlReader(MigReader):
    """
    Reads an XML file and transforms it into a Segment Group Hierarchy
    """

    _pipe_separated_names_with_digit_suffix_pattern = re.compile(
        r"^(?P<name_without_suffix>[\w\s]+)(?P<names_with_suffix>(?:\|\1\d+)+)$"
    )
    """
    https://regex101.com/r/wg5Fs5/1
    """

    def __init__(self, init_param: Union[str, Path]):
        self._original_root: etree._Element
        if isinstance(init_param, str):
            self._original_root = etree.fromstring(init_param)
            self._sanitized_root = etree.fromstring(init_param)
        elif isinstance(init_param, Path):
            self._original_root = etree.parse(str(init_param.absolute())).getroot()
            self._sanitized_root = etree.parse(str(init_param.absolute())).getroot()
        else:
            raise ValueError(f"The type of '{init_param}' is not valid")
        # self._unpack_virtual_groups() # check if this is needed at some point in the future; I don't know yet
        # the original tree is the unmodified MIG XML Structure with all its quircks
        self._original_tree: etree.ElementTree = etree.ElementTree(self._original_root)
        # but turns out, it's much easier to handle a sanitized tree that is simplified in a sense that
        # * has only lower case (ahb)names which are easy to match because they don't contain whitespace,"-" or casing
        # * has the same structure as the _original_tree so that absolute path expressions from the sanitized tree match
        self._sanitized_tree: etree.ElementTree = etree.ElementTree(self._sanitized_root)
        make_tree_names_comparable(self._sanitized_tree)

    def _unpack_virtual_groups(self) -> None:
        """
        unpacks groups with @meta.virtual="true".
        Any <root><group meta.virtual="true"><foo/><bar/></group></root> will be unpacked and modify the root to
        <root><foo/><bar/></root>
        :return:
        """
        elements_that_have_foo_groups = self._original_root.xpath('//class[@meta.virtual="true"]/parent::*')
        for element_that_has_foo_groups in elements_that_have_foo_groups:
            for foo_group in element_that_has_foo_groups.xpath('./class[@meta.virtual="true"]'):
                foo_group_index = element_that_has_foo_groups.index(foo_group)
                for foo_field in foo_group:
                    element_that_has_foo_groups.insert(foo_group_index - 1, foo_field)
                element_that_has_foo_groups.remove(foo_group)

    def get_format_name(self) -> str:
        """
        the root element of the XML is the name of the EDIFACT format
        """
        return self._original_root.tag

    def element_to_edifact_stack(self, element: etree.Element, use_sanitized_tree: bool) -> EdifactStack:
        """
        extract the edifact seed path from the given element.
        """
        # this method is directly unittests. Please refer to the test for some easy to debug examples.
        if use_sanitized_tree:
            xpath = self._sanitized_tree.getpath(element)
        else:
            xpath = self._original_tree.getpath(element)
        stack = EdifactStack(levels=[])
        iter_path = "/" + xpath.split("/")[1]
        for leaf in xpath.split("/")[2:]:
            iter_path += "/" + leaf
            # here we _always_ need to use the original root!
            leaf_element = self._original_root.xpath(iter_path)[0]  # type:ignore[attr-defined]
            level_name: str
            # todo: maybe skip virtual groups
            is_groupable = leaf_element.tag == "class"
            attribute_keys_sorted_by_priority: List[str]
            if is_groupable:
                attribute_keys_sorted_by_priority = ["migName", "ahbName", "name"]
            else:
                # I didn't create the data. I'm just trying to cope with it...
                attribute_keys_sorted_by_priority = ["ahbName", "name", "migName"]
            for attribute_key in attribute_keys_sorted_by_priority:
                if attribute_key in leaf_element.attrib:
                    level_name = leaf_element.attrib[attribute_key]
                    break
            # https://stackoverflow.com/questions/47972143/using-attr-with-pylint
            # pylint: disable=no-member
            stack.levels.append(EdifactStackLevel(name=level_name, is_groupable=is_groupable))
        return stack

    def _get_candidate_index_from_key(self, layer: AhbLocationLayer, candidates: List[Element]) -> int:
        key_set: Set[str]
        possible_results: List[int] = []
        for candidate_index, key_set in enumerate(
            set(get_nested_qualifiers("key", candidate) or []) for candidate in candidates
        ):
            if layer.opening_qualifier in key_set:
                possible_results.append(candidate_index)
        if len(possible_results) == 0:
            raise ValueError(f"Couldn't find any candidate with opening_qualifier '{layer.opening_qualifier}'")
        # todo: what if there are >1 matches. using the first one just hides data problems. we should use one instead
        return first(possible_results)

    # First make it work, then split it up
    # pylint:disable=too-many-branches
    def get_element(self, ahb_location: AhbLocation) -> Element:
        """
        Finds and returns the segment group for the specified location.
        Raises ValueErrors if it cannot find the group or the result would be ambiguous.
        """
        candidates: List[Element]
        final_query_path = f"/{self.get_format_name()}/class[@ref='/']"
        for layer in ahb_location.layers:
            query_path = final_query_path + f"/class[@ref='{layer.segment_group_key or 'UNH'}']"
            candidates = list(self._original_root.xpath(query_path))
            if len(candidates) == 0:
                raise ValueError(f"No element found for path {query_path}")
            if len(candidates) > 1:
                candidate_index = self._get_candidate_index_from_key(layer, candidates)
                final_query_path = query_path + f"[{candidate_index + 1}]"  # xpath index starts at 1, not 0
                candidates = [candidates[candidate_index]]  # list must only contain 1 remaining item at this point
                continue
            final_query_path = query_path
        if ahb_location.segment_code is not None and ahb_location.segment_code != "UNH":
            # if there is a separate class for the segment, handle it here... is most cases it's not
            query_path = final_query_path + f"/class[@ref='{ahb_location.segment_code}']"
            segment_candidates = list(self._original_root.xpath(query_path))
            if len(segment_candidates) > 1:
                for candidate_index, key_set in enumerate(
                    set(get_nested_qualifiers("key", candidate) or []) for candidate in segment_candidates
                ):
                    # if layer is undefined here, we got other problems; it's ok to crash
                    if layer.opening_qualifier in key_set:  # pylint:disable=undefined-loop-variable
                        final_query_path = query_path + f"[{candidate_index + 1}]"  # xpath index starts at 1, not 0
                        candidates = segment_candidates
                        break
            elif len(segment_candidates) == 1:
                final_query_path = query_path
        if ahb_location.data_element_id is not None:
            # now inside the remaining segment group find the entry that has the correct data element id
            query_path = final_query_path + f"/field[@meta.id='{ahb_location.data_element_id}']"  # todo:virtual groups
            candidates = list(self._original_root.xpath(query_path))
            if len(candidates) == 0:
                # todo: go a level up
                raise ValueError(f"No element found for path {query_path}")
            if len(candidates) > 1:
                if ahb_location.qualifier is not None:
                    candidate_index = one(
                        (
                            n
                            for n, c in enumerate(candidates)
                            if ahb_location.qualifier in set(get_nested_qualifiers("ref", c) or [])
                        ),
                        too_short=ValueError(f"Couldn't find any candidate with ref '{ahb_location.qualifier}'"),
                    )
                    final_query_path = query_path + f"[{candidate_index + 1}]"
                    candidates = candidates[candidate_index : candidate_index + 1]
                else:
                    raise ValueError(
                        f"Couldn't find a unique candidate with data element id '{ahb_location.data_element_id}'"
                    )
        return one(candidates)

    def get_edifact_stack(self, location: AhbLocation) -> EdifactStack:
        """
        get the edifact stack for the given segment_group, segment... combination or None if there is no match
        """
        element = self.get_element(location)
        return self.element_to_edifact_stack(element, use_sanitized_tree=False)

    def to_segment_group_hierarchy(self) -> SegmentGroupHierarchy:
        """
        convert the read data into a segment group hierarchy
        :return:
        """
        raise NotImplementedError("The inheriting class has to implement this method")



"""
Classes that allow to read XML files that contain structural information (Message Implementation Guide information)
"""
from abc import ABC, abstractmethod

from maus.models.edifact_components import EdifactStack
from maus.models.message_implementation_guide import SegmentGroupHierarchy
from maus.navigation import AhbLocation


class MigReader(ABC):
    """
    A MIG reader is a class that reads Message Implementation Guide (MIG) data from a source
    """

    @abstractmethod
    def to_segment_group_hierarchy(self) -> SegmentGroupHierarchy:
        """
        convert the read data into a segment group hierarchy
        :return:
        """
        raise NotImplementedError("The inheriting class has to implement this method")

    # pylint:disable=too-many-arguments
    @abstractmethod
    def get_edifact_stack(self, location: AhbLocation) -> EdifactStack:
        """
        Returns the edifact stack for the given combination of segment group, key, data element and name
        """
        raise NotImplementedError("The inheriting class has to implement this method")



"""
models are the data classes used to store information from e.g. ahbs.
This module contains methods available to all methods in the package.
"""


# pylint: disable=unused-argument
def _check_that_string_is_not_whitespace_or_empty(instance, attribute, value):
    """
    Check that string in the instance attribute value is not empty
    """
    if not value:
        raise ValueError(f"The string {attribute.name} must not be None or empty")
    if len(value.strip()) == 0:
        raise ValueError(f"The string {attribute.name} must not consist only of whitespace: '{value}'")



"""
module that (de)serializes AhbLocations to/from XML
"""
from typing import List, Union

import attrs

from maus.navigation import AhbLocation, AhbLocationLayer

try:
    import xmltodict
    from lxml import etree  # type:ignore[import]

    # pylint:disable=no-name-in-module
    from lxml.etree import _Element  # type:ignore[import]
except ImportError as import_error:
    import_error.msg += "; Did you install maus[xml]?"
    # lxml and xmlttodict are optional dependencies of maus but in this module, it is required
    raise


_AHB_LOCATION_LIST_XML_TAG_NAME = "ahbLocations"
_AHB_LOCATION_XML_TAG_NAME = "ahbLocation"
_AHB_LOCATION_LAYER_XML_TAG_NAME = "ahbLocationLayer"


# pylint:disable=c-extension-no-member
def _create_null_element() -> _Element:
    """Create an element with a null value using the "xsi:nil" attribute"""
    # Ideally you add
    # xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    # to the xml root element
    null_element = etree.Element("null_element", attrib={"{xsi}nil": "true"})
    return null_element


# pylint:disable=c-extension-no-member
def to_xml_element(location: Union[AhbLocation, AhbLocationLayer]) -> etree.Element:
    """
    converts an AhbLocation or AhbLocationLayer to an XML element
    """
    dict_item: dict
    if isinstance(location, dict):
        dict_item = location
        result = etree.Element(_AHB_LOCATION_LAYER_XML_TAG_NAME)
    elif isinstance(location, AhbLocation):
        dict_item = attrs.asdict(location)
        result = etree.Element(_AHB_LOCATION_XML_TAG_NAME)
    elif isinstance(location, AhbLocationLayer):
        dict_item = attrs.asdict(location)
        result = etree.Element(_AHB_LOCATION_LAYER_XML_TAG_NAME)
    else:
        raise ValueError(f"Unexpected input {location}")
    for attr_name, attr_value in dict_item.items():
        if isinstance(attr_value, list):  # it can only be a AhbLocationLayer list
            subroot = etree.SubElement(result, attr_name)
            for sub_attr in attr_value:
                sub_elem = to_xml_element(sub_attr)
                subroot.append(sub_elem)
        else:
            if attr_value is None:
                etree.SubElement(result, attr_name, attrib={"{xsi}nil": "true"})
            else:
                elem = etree.SubElement(result, attr_name)
                elem.text = str(attr_value)
    return result


def to_xml_elements(locations: List[AhbLocation]) -> etree.Element:
    """
    converts multiple AhbLocations to an XML element
    """
    result = etree.Element(_AHB_LOCATION_LIST_XML_TAG_NAME)
    for location in locations:
        location_element = to_xml_element(location)
        result.append(location_element)
    return result


def _replace_nil_element_with_none(raw_dict: dict) -> dict:
    sanitized_values: dict = {}
    for key, value in raw_dict.items():
        if isinstance(value, dict) and any(
            _key.startswith("@ns") and _key.endswith(":nil") and _value == "true" for _key, _value in value.items()
        ):
            sanitized_values[key] = None
        else:
            sanitized_values[key] = value
    return sanitized_values


def _xml_dict_to_location(xml_dict: dict) -> AhbLocation:
    layers: List[AhbLocationLayer] = []
    layer_dicts = xml_dict["layers"][_AHB_LOCATION_LAYER_XML_TAG_NAME]
    for ahb_location_layer_dict in layer_dicts:
        sanitized_values = _replace_nil_element_with_none(ahb_location_layer_dict)
        layer = AhbLocationLayer(**sanitized_values)
        layers.append(layer)
    del xml_dict["layers"]
    sanitized_leftovers = _replace_nil_element_with_none(xml_dict)
    return AhbLocation(layers=layers, **sanitized_leftovers)


# pylint:disable=c-extension-no-member
def from_xml_element(xml: Union[str, _Element]) -> AhbLocation:
    """
    deserializes an AhbLocation from an etree element or str
    :param xml:
    :return:
    """
    if isinstance(xml, _Element):
        xml_str = etree.tostring(xml)
    else:
        xml_str = xml
    xml_dict = xmltodict.parse(xml_str)[_AHB_LOCATION_XML_TAG_NAME]
    return _xml_dict_to_location(xml_dict)


def from_xml_elements(xml: Union[str, _Element]) -> List[AhbLocation]:
    """
    deserializes multiple ahb locations
    """
    if isinstance(xml, _Element):
        xml_str = etree.tostring(xml)
    else:
        xml_str = xml
    xml_list = xmltodict.parse(xml_str)[_AHB_LOCATION_LIST_XML_TAG_NAME]
    result = []
    element = xml_list[_AHB_LOCATION_XML_TAG_NAME]
    if isinstance(element, list):
        for item in element:
            sub_result = _xml_dict_to_location(item)
            result.append(sub_result)
        return result
    if isinstance(element, dict):
        return [_xml_dict_to_location(element)]
    raise ValueError("unexpected input")



"""
This script is run in the tox 'json_schemas' environment.
It creates json schema files as described in the README.md in the same directory.
"""

import json
import pathlib
from typing import List, Type

from marshmallow import Schema
from marshmallow_jsonschema import JSONSchema  # type:ignore[import]

from maus.models.anwendungshandbuch import DeepAnwendungshandbuchSchema

schema_types: List[Type[Schema]] = [
    DeepAnwendungshandbuchSchema
    # add other schemas as you like
]
json_schema = JSONSchema()
for schema_type in schema_types:
    this_directory = pathlib.Path(__file__).parent.absolute()
    file_name = schema_type.__name__ + ".json"  # pylint:disable=invalid-name
    file_path = this_directory / file_name
    schema_instance = schema_type()
    json_schema_dict = json_schema.dump(schema_instance)

    with open(file_path, "w", encoding="utf-8") as json_schema_file:
        json.dump(json_schema_dict, json_schema_file, ensure_ascii=False, sort_keys=True, indent=4)



"""
This module contains (static) functions that process single lxml.etree Elements.
Each function is separately unit tested.
"""
import re
from typing import List, Literal, Optional

# pylint:disable=no-name-in-module
from lxml.etree import Element  # type:ignore[import]


def get_segment_group_key_or_none(element: Element) -> Optional[str]:
    """
    returns the segment group of element if present; None otherwise
    """
    if "ref" in element.attrib and element.attrib["ref"].startswith("SG"):
        # the trivial case
        return element.attrib["ref"]
    return None


#: a regex to match a ref-segment: https://regex101.com/r/D81bbO/1
_single_nested_qualifier_pattern = re.compile(
    r"^(?P<segment_code>[A-Z]+):\d+:(?:\d+|\(\d+,\d+\))\[(?:\w+:)+\w+:?=(?P<qualifier>[A-Z\d]+)\]$"
)

#: a regex to match multiple ref/key segments: https://regex101.com/r/6XooRL/2
_multiple_nestes_qualifiers_pattern = re.compile(
    r"(?P<segment_code>[A-Z]+):\d+:(?:\d+|\(\d+,\d+\))\[(?P<inner>(?:(?:[A-Z]+:)?\d+:\d+=[A-Z\d]+\|?)+)\]$"
)


def get_nested_qualifiers(attrib_key: Literal["ref", "key"], element: Element) -> Optional[List[str]]:
    """
    returns the nested qualifier of an element if present; None otherwise
    we still return None instead of an empty list, because all the framework around this method actually check for None
    instead of empty lists
    """
    if attrib_key in element.attrib:
        body: str = element.attrib[attrib_key]
        single_match = _single_nested_qualifier_pattern.match(body)
        if single_match:
            return [single_match["qualifier"]]
        multi_match = _multiple_nestes_qualifiers_pattern.match(body)
        if multi_match:
            # if body == "QTY:1:1[1:0=265|1:0=Z10|1:0=Z08]"
            # then multi_match["inner"] is "1:0=265|1:0=Z10|1:0=Z08"
            return [expression.split("=")[1] for expression in multi_match["inner"].split("|")]
    return None


def get_ahb_name_or_none(element: Element) -> Optional[str]:
    """
    returns the ahbName of element if present; None otherwise
    """
    if "ahbName" in element.attrib:
        return element.attrib["ahbName"]
    return None



"""
The transformation module is useful when you handle edifact transactions in both a domain/data model similar to edifact
(the "edifact domain") and a domain/data model which is used by your application (we simply call it the "non-edifact" or
"application domain").

Usually the data model similar to EDIFACT is something structurally equivalent to EDIFACT, think e.g. of a JSON or XML.
Wherever there is the possibility to group something in the form of repeated segment groups, you usually have something
like an array in your edifact-equivalent data structure. We call this structure an "EdiSeed", because it is what you
"seed" into your EdifactBuilder to generate EDIFACT. The transformation from an EdiSeed to EDIFACT is quite simple.
It's a plain mapping logic that doesn't care about cardinalities and more complicated stuff than simple replacements
and loops. Neighbouring segments in EDIFACT are transformed into neighbouring fields in the EdiSeed structure.
Think of the EdiSeed as an intermediate data structure.

MAUS is agnostic regarding your EdiSeed. It doesn't care about how your EdiSeed looks like exactly. It just needs to be
accessible somehow.

In your application you might use a different data model which is de-coupled from EDIFACT and thus also de-coupled from
the EdiSeed. This is generally a good idea, because it spares your application from the never-ending pain introduced by
BDEW and edi@energy every 6 months. However, this comes with the price of having to write a more complex transformation.

This module contains methods that are relevant when transforming data between your edi-similar structure ("EdiSeed")
and the application domain ("non-edifact").
"""
import asyncio
from typing import Any, Awaitable, Dict, List, Mapping, Optional, Protocol, TypeVar

from maus.models.anwendungshandbuch import DataElementValuePool, DeepAnwendungshandbuch
from maus.models.edifact_components import ValuePoolEntry

EdifactData = TypeVar("EdifactData")
"""
the type which you use to model the edifact side of things (e.g. the type of the EdiSeed)
"""
ApplicationData = TypeVar("ApplicationData")
"""
the type which you use to model the data in your application (e.g. the type of your BO4E, BOneyComb)
"""


class ApplicationEdifactConverter(Protocol[EdifactData, ApplicationData]):
    """
    this class describes a converter from the Application Domain to the EDIFACT Domain
    """

    async def transform_application_to_edi_domain(self, application_data: ApplicationData) -> EdifactData:
        """
        Transforms the given application data (e.g. bo4e/boneycomb) into the edifact domain (e.g. ediseed)
        :param application_data:
        :return: the converted data
        """

    async def transform_edi_to_application_domain(self, edifact_data: EdifactData) -> ApplicationData:
        """
        Transforms the given edifact data (e.g. ediseed) into the application domain (e.g. bo4e/boneycomb)
        :param edifact_data:
        :return: the converted data
        """


DataModel_contra = TypeVar("DataModel_contra", contravariant=True)


class Accessor(Protocol[DataModel_contra]):
    """description of a class that allows to get and set values from a generic data model"""

    def get_value(self, data: DataModel_contra, discriminator: str) -> Optional[Any]:
        """
        Gets the value described by the discriminator from the given data
        :param data: the data to be read
        :param discriminator: describes _which_ value should be read and returned
        :return: the value or None
        """

    def set_value(self, data: DataModel_contra, discriminator: str, value: Any) -> None:
        """
        Sets the value that is described by the disciminator in the given data instance.
        :param value: the value to be set
        :param data: the data to be modified
        :param discriminator: describes _which_ value should be set
        :return: nothing; it modifies the instance of data
        """


EdifactAccessor = Accessor[EdifactData]  #: a class that allows to access (read/write) in the edifact data model
ApplicationAccessor = Accessor[ApplicationData]  #: a class that allows to access (read/write) in the application data


# I feel like there's no way to boil this down to fewer arguments (6/5) without tradeoffs in readability.
# pylint:disable=too-many-arguments
async def generate_value_pool_replacement(
    application_data_model: ApplicationData,
    data_element: DataElementValuePool,
    converter: ApplicationEdifactConverter,
    edifact_accessor: EdifactAccessor,
    application_accessor: ApplicationAccessor,
    edifact_to_non_edifact_path_mapping: Mapping[str, str],
) -> Dict[str, str]:
    """
    generate the value pool replacement dict for the given data_element
    :param application_data_model: the application representation of the data
    :param data_element: the data_element for which the replacement shall be computed
    :param converter: a converter between application and edifact domain
    :param application_accessor: Allows to modify and access a given application data model
    :param edifact_accessor: Allows to modify and access a given edifact data model
    :param edifact_to_non_edifact_path_mapping: maps application domain (keys) to edifact domain (values) discriminators
    :return: a edifact (key) to non-edifact (value) mapping of values from the given data element value pool
    """
    initial_edifact_data = await converter.transform_application_to_edi_domain(application_data_model)
    edi_to_non_edi_value_mapping: Dict[str, str] = {}
    for edi_path, non_edi_path in edifact_to_non_edifact_path_mapping.items():
        if edi_path != data_element.discriminator:
            continue
        non_edi_value = application_accessor.get_value(application_data_model, non_edi_path)
        edi_value = edifact_accessor.get_value(
            initial_edifact_data, edi_path
        )  # this is the current value from the initial data
        if edi_value is not None:
            edi_to_non_edi_value_mapping.update({edi_value: non_edi_value})
        break
    else:
        raise ValueError(f"No mapping found for '{data_element.discriminator}'")

    # now loop over the other values
    async def _get_value_pool_mapping(value_pool_entry: ValuePoolEntry):
        # modified_edifact_data = initial_edifact_data.copy() #only works on dicts
        modified_edifact_data = initial_edifact_data  # ⚠ this modifies the reference!
        modified_edi_value = value_pool_entry.qualifier
        edifact_accessor.set_value(modified_edifact_data, edi_path, value_pool_entry.qualifier)
        # now we transform the modified edifact data back to the application domain
        modified_application_data = await converter.transform_edi_to_application_domain(modified_edifact_data)
        modified_application_value = application_accessor.get_value(modified_application_data, non_edi_path)
        return {modified_edi_value: modified_application_value}

    value_pool_mapping_tasks: List[Awaitable[Dict[str, Optional[str]]]] = [
        _get_value_pool_mapping(value_pool_entry) for value_pool_entry in data_element.value_pool
    ]
    for result in await asyncio.gather(*value_pool_mapping_tasks):
        edi_to_non_edi_value_mapping.update(result)
    return edi_to_non_edi_value_mapping


async def transform_all_value_pools(
    application_data_model: ApplicationData,
    deep_ahb: DeepAnwendungshandbuch,
    converter: ApplicationEdifactConverter,
    edifact_accessor: EdifactAccessor,
    application_accessor: ApplicationAccessor,
    edifact_to_non_edifact_path_mapping: Mapping[str, str],
) -> None:
    """
    transforms all value pools in the given deep ahb
    """
    replacement_tasks: List[Awaitable] = []

    async def transform_value_pool(value_pool: DataElementValuePool):
        replacement = await generate_value_pool_replacement(
            application_data_model=application_data_model,
            data_element=value_pool,
            converter=converter,
            edifact_accessor=edifact_accessor,
            application_accessor=application_accessor,
            edifact_to_non_edifact_path_mapping=edifact_to_non_edifact_path_mapping,
        )
        value_pool.replace_value_pool(replacement)

    for value_pool in deep_ahb.get_all_value_pools():
        # ignore Absender and Empfänger and all paths that are no json paths
        if (
            value_pool.discriminator is not None
            and value_pool.discriminator.startswith("$")
            and "Absender" not in value_pool.discriminator
            and "Empfänger" not in value_pool.discriminator
        ):
            replacement_tasks.append(transform_value_pool(value_pool))
    await asyncio.gather(*replacement_tasks)



# pylint:disable=too-few-public-methods
"""
Contains classes to Model Message Implementation Guides (MIG)
"""
from typing import List, Optional, Set, Tuple

import attrs
from marshmallow import Schema, fields, post_load

# TODO: the SegmentGroupHierarchy shall not be hardcoded but instead loaded from .tree files
# .tree files are a complete representation of an EDIFACT format including all segments and allowed data element values.
# A LARK grammar to parse them is work in progress at the moment and will be added later.
# From a parsed .tree file it is possible to also derive the segment group hierarchy.


@attrs.define(auto_attribs=True, kw_only=True, frozen=True)
class SegmentGroupHierarchy:
    """
    Models the hierarchy of segment groups within an EDIFACT format.
    Therefore each segment group is modeled together with its respective first segment inside the group.
    This is not a full representation of the EDIFACT formats structure but enough to understand the basic hierarchy.
    The Segment Group Hierarchy (SGH) is what allows us to transform a flat AHB into a NestedAhb that has a structure.
    This information / the SGH can be (manually) extracted form the diagrams in the MIG PDFs or generated from another
    source (which the BDEW does not provide, sadly).
    See the unit tests for working examples for UTILMD and MSCONS.
    It'll make sense then.
    """

    segment_group: Optional[str] = attrs.field(validator=attrs.validators.optional(attrs.validators.instance_of(str)))
    """segment group name, e.g. "SG4" or "SG5" """
    opening_segment: str = attrs.field(validator=attrs.validators.instance_of(str))
    """first segment in group, e.g. 'IDE' or 'LOC' """
    sub_hierarchy: Optional[
        List["SegmentGroupHierarchy"]
    ]  #: segment groups below this level, e.g.[(SG5, LOG), (SG6,RFF), (SG8, SEQ), ...]

    def flattened(self) -> List[Tuple[Optional[str], str]]:
        """
        Returns the segment groups and opening tuples in the order in which they are expected to occur in an AHB.
        This is useful to match this deep structure with a flat AHB.
        """
        result: List[Tuple[Optional[str], str]] = [(self.segment_group, self.opening_segment)]
        if self.sub_hierarchy is not None:
            for sub_hier in self.sub_hierarchy:
                result += sub_hier.flattened()
        return result

    def is_hierarchically_below(self, segment_group_key: Optional[str]) -> bool:
        """
        returns true iff the segment_group provided is a (direct or indirect) sub group of self.
        """
        if self.sub_hierarchy is None:
            return False
        for sub_hier in self.sub_hierarchy:
            if sub_hier.segment_group == segment_group_key:
                return True
            if sub_hier.is_hierarchically_below(segment_group_key):
                return True
        return False

    def sg_is_hierarchically_below(
        self, segment_group_key_x: Optional[str], segment_group_key_y: Optional[str]
    ) -> bool:
        """
        returns true iff segment group with segment_group_key_x is hierarchically below the segment_group with key_y
        :param segment_group_key_x:
        :param segment_group_key_y:
        :return:
        """
        if segment_group_key_x == segment_group_key_y:
            return False
        if self.segment_group == segment_group_key_y:
            # because, if segment_group_x exists, it will be hierarchically below the root (y)
            all_sg_keys_in_this_sgh: Set[Optional[str]] = {x[0] for x in self.flattened()}
            if segment_group_key_x in all_sg_keys_in_this_sgh:
                return True
        for sub_hierarchy in self.sub_hierarchy or []:
            if sub_hierarchy.sg_is_hierarchically_below(segment_group_key_x, segment_group_key_y) is True:
                return True
        return False


class SegmentGroupHierarchySchema(Schema):
    """
    A schema to (de-)serialize :class:`.SegmentGroupHierarchy`
    """

    segment_group = fields.String(required=True, allow_none=True)
    opening_segment = fields.String(required=True)
    # disable unnecessary lambda warning because of circular imports
    sub_hierarchy = fields.List(
        fields.Nested(lambda: SegmentGroupHierarchySchema()), allow_none=True  # pylint:disable=unnecessary-lambda
    )

    # pylint:disable=unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> SegmentGroupHierarchy:
        """
        Converts the barely typed data dictionary into an actual :class:`.SegmentGroupHierarchy`
        """
        return SegmentGroupHierarchy(**data)



"""
maus is the MIG AHB Utility Stack
"""
import json
from pathlib import Path

import click

from maus.mig_ahb_matching import to_deep_ahb
from maus.models.anwendungshandbuch import (
    DeepAnwendungshandbuch,
    DeepAnwendungshandbuchSchema,
    FlatAnwendungshandbuchSchema,
)
from maus.models.message_implementation_guide import SegmentGroupHierarchySchema
from maus.reader.mig_xml_reader import MigXmlReader


# pylint:disable=too-many-arguments
@click.command()
@click.version_option()
@click.option(
    "-fap",
    "--flat_ahb_path",
    type=click.Path(exists=True, path_type=Path),
    help="Path to the flat ahb json file",
    required=True,
)
@click.option(
    "-sghp",
    "--sgh_path",
    type=click.Path(exists=True, path_type=Path),
    help="Path to the sgh json file",
    required=True,
)
@click.option(
    "-tp",
    "--template_path",
    type=click.Path(exists=True, path_type=Path),
    help="Path to the template file",
    required=True,
)
@click.option(
    "-cp",
    "--check_path",
    type=click.Path(exists=True, path_type=Path),
    help="Path to the maus json file",
)
@click.option(
    "-o",
    "--output_path",
    type=click.Path(dir_okay=False, file_okay=True, path_type=Path),
    help="Path to the output file",
)
# pylint:disable=no-value-for-parameter
def main(
    flat_ahb_path: Path,
    sgh_path: Path,
    template_path: Path,
    check_path: Path,
    output_path: Path,
):
    """
    🐭 MAUS CLI is a standalone executable that generates .maus.json files from given input data
    """

    if check_path is None and output_path is None or check_path is not None and output_path is not None:
        # pylint:disable=line-too-long
        click.secho(
            "❌ You need to specify either the `output_path` or the `check_path` parameter. Please use --help to see more information.",
            fg="red",
        )
        raise click.Abort()

    with open(flat_ahb_path, "r", encoding="utf-8") as flat_ahb_file:
        flat_ahb = FlatAnwendungshandbuchSchema().load(json.load(flat_ahb_file))
    with open(sgh_path, "r", encoding="utf-8") as sgh_file:
        sgh = SegmentGroupHierarchySchema().loads(sgh_file.read())

    mig_reader = MigXmlReader(template_path)

    # create new maus.json files
    maus = to_deep_ahb(flat_ahb, sgh, mig_reader)

    if output_path is not None and check_path is not None:
        click.secho("❌ You can only specify one of the output_path and maus_to_check_path parameters", fg="red")
        raise click.Abort()

    if output_path is not None:
        maus_dict = DeepAnwendungshandbuchSchema().dump(maus)

        with open(output_path, "w", encoding="utf-8") as maus_file:
            json.dump(maus_dict, maus_file, indent=2, ensure_ascii=False, sort_keys=True)

    if check_path is not None:
        with open(check_path, "r", encoding="utf-8") as maus_file:
            expected_maus: DeepAnwendungshandbuch = DeepAnwendungshandbuchSchema().loads(maus_file.read())

            # reset the line index to make the comparison work
            # this is fine cause there is no logic built on top of the line index
            maus.reset_ahb_line_index()
            expected_maus.reset_ahb_line_index()

            if expected_maus == maus:
                click.secho("✅ The generated maus.json matches the expected one", fg="green")
            else:
                click.secho("❌ The generated maus.json does not match the expected one!", fg="red")
                raise click.Abort()


if __name__ == "__main__":
    main()  # pylint:disable=no-value-for-parameter



"""
This module contains helper methods that compare strings, especially names from AHB and MIG.
"""
from typing import Literal, Optional, overload

from lxml import etree  # type:ignore[import]


@overload
def make_name_comparable(orig_str: str) -> str:
    ...


@overload
def make_name_comparable(orig_str: Literal[None]) -> Literal[None]:
    ...


def make_name_comparable(orig_str: Optional[str]) -> Optional[str]:
    """
    Removes all the characters that could be a problem when matching names from the AHB with names from the MIG
    """
    if orig_str is None:
        return None
    result: str = orig_str.lower()
    result = result.replace("ß", "ss")
    for removable_character in [" ", "-", "\n"]:
        result = result.replace(removable_character, "")
    return result


def are_similar_names(name_x: Optional[str], name_y: Optional[str]) -> bool:
    """
    Returns true if name_x and name_y are somehow similar.
    "Somehow similar" in this context means, that all the artefacts from text to word to PDF + scraping
    + human errors might add up to something which explains the difference between name_x and name_y.
    """
    if (name_x and not name_y) or (name_y and not name_x):
        return False
    if (not name_x) and (not name_y):
        return True
    # neither name_x nor name_y are None below this line
    return make_name_comparable(name_x) == make_name_comparable(name_y)  # type:ignore[arg-type]


def make_tree_names_comparable(tree: etree.ElementTree) -> None:  # pylint:disable=c-extension-no-member
    """
    modifies the provided tree by applying `make_name_comparable` to all name and ahbName attributes
    """
    for element in tree.iter():
        for attrib_key, attrib_value in list(element.attrib.items()):
            if attrib_key in {"name", "ahbName"}:
                element.attrib[attrib_key] = make_name_comparable(attrib_value)



"""
This module manages EDIFACT related stuff. It's basically a helper module to avoid stringly typed parameters.
"""
import datetime
import re
from enum import Enum
from typing import Dict, Optional

import attrs

_PRUEFI_REGEX = r"^[1-9]\d{4}$"
pruefidentifikator_pattern = re.compile(_PRUEFI_REGEX)


# pylint: disable=too-few-public-methods
class EdifactFormat(str, Enum):
    """
    existing EDIFACT formats
    """

    APERAK = "APERAK"
    COMDIS = "COMDIS"  #: communication dispute
    CONTRL = "CONTRL"  #: control messages
    IFTSTA = "IFTSTA"  #: Multimodaler Statusbericht
    INSRPT = "INSRPT"  #: Prüfbericht
    INVOIC = "INVOIC"  #: invoice
    MSCONS = "MSCONS"  #: meter readings
    ORDCHG = "ORDCHG"  #: changing an order
    ORDERS = "ORDERS"  #: orders
    ORDRSP = "ORDRSP"  #: orders response
    PRICAT = "PRICAT"  #: price catalogue
    QUOTES = "QUOTES"  #: quotes
    REMADV = "REMADV"  #: zahlungsavis
    REQOTE = "REQOTE"  #: request quote
    PARTIN = "PARTIN"  #: market partner data
    UTILMD = "UTILMD"  #: utilities master data
    UTILTS = "UTILTS"  #: formula

    def __str__(self):
        return self.value


_edifact_mapping: Dict[str, EdifactFormat] = {
    "99": EdifactFormat.APERAK,
    "29": EdifactFormat.COMDIS,
    "21": EdifactFormat.IFTSTA,
    "23": EdifactFormat.INSRPT,
    "31": EdifactFormat.INVOIC,
    "13": EdifactFormat.MSCONS,
    "39": EdifactFormat.ORDCHG,
    "17": EdifactFormat.ORDERS,
    "19": EdifactFormat.ORDRSP,
    "27": EdifactFormat.PRICAT,
    "15": EdifactFormat.QUOTES,
    "33": EdifactFormat.REMADV,
    "35": EdifactFormat.REQOTE,
    "37": EdifactFormat.PARTIN,
    "11": EdifactFormat.UTILMD,
    "25": EdifactFormat.UTILTS,
    "91": EdifactFormat.CONTRL,
    "92": EdifactFormat.APERAK,
    "44": EdifactFormat.UTILMD,  # UTILMD for GAS since FV2310
    "55": EdifactFormat.UTILMD,  # UTILMD for STROM since FV2310
}


class EdifactFormatVersion(str, Enum):
    """
    One format version refers to the period in which an AHB is valid.
    """

    FV2104 = "FV2104"  #: valid from 2021-04-01 until 2021-10-01
    FV2110 = "FV2110"  #: valid from 2021-10-01 until 2022-04-01
    FV2210 = "FV2210"  #: valid from 2022-10-01 onwards ("MaKo 2022", was 2204 previously)
    FV2304 = "FV2304"  #: valid from 2023-04-01 onwards
    # whenever you add another value here, please also make sure to add its key date to get_edifact_format_version below

    def __str__(self):
        return self.value


def get_edifact_format_version(key_date: datetime.datetime) -> EdifactFormatVersion:
    """
    :return: the edifact format version that is valid on the specified key date
    """
    if key_date < datetime.datetime(2021, 9, 30, 22, 0, 0, 0, tzinfo=datetime.timezone.utc):
        return EdifactFormatVersion.FV2104
    if key_date < datetime.datetime(2022, 9, 30, 22, 0, 0, 0, tzinfo=datetime.timezone.utc):
        return EdifactFormatVersion.FV2110
    if key_date < datetime.datetime(2023, 3, 31, 22, 0, 0, 0, tzinfo=datetime.timezone.utc):
        return EdifactFormatVersion.FV2210
    return EdifactFormatVersion.FV2304


def get_current_edifact_format_version() -> EdifactFormatVersion:
    """
    returns the edifact_format_version that is valid as of now
    """
    tz_aware_now = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)
    return get_edifact_format_version(tz_aware_now)


def is_edifact_boilerplate(segment_code: Optional[str]) -> bool:
    """
    returns true iff this segment is not relevant in a sense that it has to be validated or merged with the AHB
    """
    if not segment_code:
        return True
    return segment_code.strip() in {"UNT", "UNZ"}


def get_format_of_pruefidentifikator(pruefidentifikator: str) -> EdifactFormat:
    """
    returns the format corresponding to a given pruefi
    """
    if not pruefidentifikator:
        raise ValueError("The pruefidentifikator must not be falsy")
    if not pruefidentifikator_pattern.match(pruefidentifikator):
        raise ValueError(f"The pruefidentifikator '{pruefidentifikator}' is invalid.")
    try:
        return _edifact_mapping[pruefidentifikator[:2]]
    except KeyError as key_error:
        raise ValueError(f"No Edifact format was found for pruefidentifikator '{pruefidentifikator}'.") from key_error


# pylint:disable=unused-argument
def _check_that_pruefi_and_format_are_consistent(instance: "EdiMetaData", attribute, value: str):
    """
    The value is the pruefidentifikator, the instance is the EdiMetaData instance.
    This validator raises an ValueError if the pruefidentifikator is not consistent with the instance.edifact_format.
    """
    actual_format = instance.edifact_format
    expected_format = get_format_of_pruefidentifikator(value)
    if actual_format != expected_format:
        raise ValueError(f"'{value}' is incompatible with '{actual_format}'; expected '{expected_format}' instead")


@attrs.define(kw_only=True, frozen=True, auto_attribs=True)
class EdiMetaData:
    """
    a container that contains edifact-related metadata
    """

    pruefidentifikator: str = attrs.field(
        validator=attrs.validators.and_(
            attrs.validators.instance_of(str),
            attrs.validators.matches_re(_PRUEFI_REGEX),
            _check_that_pruefi_and_format_are_consistent,
        )
    )
    """
    The pruefidentifikator, e.g. '11042'
    """
    edifact_format: EdifactFormat = attrs.field(validator=attrs.validators.instance_of(EdifactFormat))
    """
    The Edifact Format, e.g. 'UTILMD'
    """

    @edifact_format.default
    def _get_format_from_pruefidentifikator(self):
        return get_format_of_pruefidentifikator(self.pruefidentifikator)

    edifact_format_version: EdifactFormatVersion = attrs.field(
        validator=attrs.validators.instance_of(EdifactFormatVersion)
    )
    """
    The Edifact Format Version, e.g. 'FV2210'
    """



"""
A module to parse Hochfrequenz .tree files and return them as Segment Group Hierarchy
"""
from pathlib import Path
from typing import Union

from more_itertools import first

try:
    from lark import Lark, Tree
except ImportError as import_error:
    import_error.msg += "; Did you install maus[tree]?"
    # lark is only an optional dependency of maus but in this module, it is required
    raise

GRAMMAR = r"""
?start:SEGMENT_GROUP_CHAIN
SEGMENT_GROUP_CHAIN: ROOT_NODE ":" SEGMENT_GROUP ("\n" SEGMENT_GROUP)*
SEGMENT_GROUP: ((SEGMENT_GROUP_KEY) ":"?)? SEGMENT ("," (SEGMENT|SEGMENT_GROUP_REFERENCE))* // e.g. "SG4:CTA[M;M;C*Can3+C*Can17*Can256;C*Ran3{IC}+R*N*Ran256],COM[C;R;M*Man512*Man3;M*Man512*Man3{EM.FX.TE.AJ.AL}]"
SEGMENT_GROUP_REFERENCE: SEGMENT_GROUP_KEY "[" OBLIG_MARK ";" OBLIG_MARK "]"
SEGMENT: SEGMENT_KEY "[" INTERNATIONAL_BDEW_TUPLE_CHAIN "]" // e.g. "CTA[M;M;C*Can3+C*Can17*Can256;C*Ran3{IC}+R*N*Ran256]"
INTERNATIONAL_BDEW_TUPLE_CHAIN: INTERNATIONAL_BDEW_TUPLE (";" INTERNATIONAL_BDEW_TUPLE)* // e.g. "C;R;M*Man512*Man3;M*Man512*Man3{EM.FX.TE.AJ.AL}"
INTERNATIONAL_BDEW_TUPLE: (CONDITIONED_DATA_ELEMENTS_SWITCH*|MULTIPLE_DATA_ELEMENTS";") (CONDITIONED_DATA_ELEMENTS_SWITCH*|MULTIPLE_DATA_ELEMENTS) // f.e. "C*Can3+C*Can17*Can256;C*Ran3{IC}"
CONDITIONED_DATA_ELEMENTS_SWITCH: CONDITION MULTIPLE_DATA_ELEMENTS ["|"]
CONDITION: "(" CONDITION_BODY ("|"CONDITION_BODY)* ")" // e.g. "(LOC#1#0=237)" but als (LOC#1#0=237|FOO#2#3=BAR)
CONDITION_BODY: SEGMENT_KEY "#" INT "#"INT "!"?"=" VALUE_POOL_ENTRY // "LOC#1#0=237"
SEGMENT_KEY: /[A-Z]{3,}/ // e.g. "NAD"
MULTIPLE_DATA_ELEMENTS: COMPOSITE_CHAIN ("+" COMPOSITE_CHAIN)* // e.g. "C*Can3+C*Can17*Can256"
COMPOSITE_CHAIN: COMPOSITE ("*" COMPOSITE)* // e.g. "C*Ran3{IC}"
COMPOSITE: OBLIG_MARK [CHARACTER_SET [COMPOSITE_LENGTH [VALUE_POOL]]] // e.g. "Ran3{IC}" but also plain "M"
VALUE_POOL: "{" VALUE_POOL_ENTRY("." VALUE_POOL_ENTRY)*  "}" // e.g. "{EM.FX.TE.AJ.AL}"
VALUE_POOL_ENTRY: /[A-Za-z\d\-_]+/ // e.g. "E01" // lower case and "-" for "Gabi-RLMoT" or "H2_ENGY"
COMPOSITE_LENGTH: INT // e.g. 35
CHARACTER_SET: "an"|"nn"|"np"|"aa"
OBLIG_MARK: "M"|"R"|"N"|"C"|"D"|"O" // see issue 11
ROOT_NODE: "/"
SEGMENT_GROUP_KEY: "SG" INT|"UNH"|"UNB" // e.g. "SG12"
%import common.INT
%import common.WS
%ignore WS  // WS = whitespace
"""
_parser = Lark(GRAMMAR)


def read_tree(tree: Union[str, Path]) -> Tree:
    """
    tries to parse the given tree using our .tree grammar.
    :param tree: string with the .tree content or the path to the tree file
    :return: a parsed tree
    """
    tree_str: str
    if isinstance(tree, Path):
        with open(tree, "r", encoding="utf-8") as tree_file:
            tree_str = tree_file.read()
    elif isinstance(tree, str):
        tree_str = tree
    else:
        raise ValueError(f"The provided argument is neither a path nor a str but a {tree.__class__}")
    parsed_tree = _parser.parse(tree_str)
    return parsed_tree


# todo: This parsed tree is not useful yet, but it's the basis for any future analysis/extraction.


def check_file_can_be_parsed_as_tree(file_path: Path) -> None:
    """
    Returns nothing iff the given file is parsable as .tree and contains no obvious errors.
    In case of error an exception is raised.
    """
    _ = read_tree(file_path)


def check_youngest_tree_is_parseable(dir_path: Path) -> None:
    """
    applies check_file_can_be_parsed_as_tree to the latest .tree file in the given directory
    :param dir_path:
    :return: nothing
    """
    if not dir_path.exists():
        return
    if not dir_path.is_dir():
        raise ValueError(f"You must only provide a directory path to this function but '{dir_path}' is not")
    tree_files_in_directory = sorted(dir_path.glob("*.tree"), reverse=True)
    if len(tree_files_in_directory) == 0:
        return
    relevant_file_to_check = first(tree_files_in_directory)
    check_file_can_be_parsed_as_tree(relevant_file_to_check)



"""
the navigation module is contains models and code that allow to "navigate" through the AHB and MIG structure.
I.e. it allows to loop over an Anwendungshandbuch and "remember" which turns we took in the MIG structure (each turn is
a AhbLocationLayer) in order to arrive at a certain line of the AHB. This information is stored in an AhbLocation.
"""
import sys
from enum import Enum
from typing import Callable, List, Optional, Tuple, TypeVar, Union, overload

import attrs
from more_itertools import first_true, last

from maus.models.anwendungshandbuch import AhbLine
from maus.models.message_implementation_guide import SegmentGroupHierarchy

T = TypeVar("T")


def _enhance_with_next_line_that_fulfills_predicate(
    ahb_lines: List[AhbLine], predicate: Callable[[AhbLine], bool], selector: Callable[[AhbLine], T]
) -> List[Tuple[AhbLine, Optional[T]]]:
    """
    For each line in ahb_line, find the next line that fulfills the predicate and return its property described by the
    selector.
    """
    # this function is not unit tested directly but only via _enhance_with_segment and _enhance_with_first_qualifier
    result: List[Tuple[AhbLine, Optional[T]]] = []
    for index, ahb_line in enumerate(ahb_lines):
        if predicate(ahb_line):
            result.append((ahb_line, selector(ahb_line)))
        else:
            first_line_that_matches_predicate = first_true(ahb_lines[index:], pred=predicate)
            if first_line_that_matches_predicate is None:
                break
            result.append((ahb_line, selector(first_line_that_matches_predicate)))
    ahb_line_index = len(result) - 1
    while len(result) < len(ahb_lines):
        result.append((ahb_lines[ahb_line_index], None))
        ahb_line_index += 1
    return result


def _enhance_with_next_segment(ahb_lines: List[AhbLine]) -> List[Tuple[AhbLine, str]]:
    """
    loop over the given ahb_lines, and return the same (unmodified) lines + the for each line the next segment code,
    which is not none. Here's an example:
    lines:
    [
    (line: Segment Group, Segment, Data Element)
    {SG1, None, None}
    {SG1, ABC, None}
    {SG1, ABC, 1234}
    {SG2, None, None}
    {SG2, DEF, None}
    {SG2, DEF, 5678}
    ]
    will return:
    [
    (line, str)
    {SG1, None, None}, ABC
    {SG1, ABC, None}, ABC
    {SG1, ABC, 1234}, ABC
    {SG2, None, None}, DEF
    {SG2, DEF, None}, DEF
    {SG2, DEF, 5678}, DEF
    ]
    """

    return _enhance_with_next_line_that_fulfills_predicate(
        ahb_lines=ahb_lines,
        predicate=lambda line: line.segment_code is not None,
        selector=lambda line: line.segment_code,  # type:ignore[return-value, arg-type]
        # predicate ensures that the selector does _not_ return None
    )


def _enhance_with_next_value_pool_entry(ahb_lines: List[AhbLine]) -> List[Tuple[AhbLine, str]]:
    """
    loop over the given ahb_lines, and return the same (unmodified) lines + the for each line the next value pool entry,
    which is not none. This is relevant to find the important first qualifier of the first segment in each segment group
    Here's an example:
    lines:
    [
    (line: Segment Group, Segment, Data Element, ValuePoolEntry)
    {SG1, ABC, None, None}
    {SG2, ABC, 1234, Z01}
    {SG2, DEF, None, None}
    {SG2, DEF, 5678, Z02}
    ]
    will return:
    [
    (line, str)
    {SG1, ABC, None, None}, Z01
    {SG1, ABC, 1234, Z01}, Z01
    {SG2, DEF, None, None}, Z02
    {SG2, DEF, 5678, Z02}, Z02
    ]
    """

    return _enhance_with_next_line_that_fulfills_predicate(
        ahb_lines=ahb_lines,
        predicate=lambda line: line.value_pool_entry is not None,
        selector=lambda line: line.value_pool_entry,  # type:ignore[return-value, arg-type]
        # predicate ensures that the selector does _not_ return None
    )


def _is_opening_segment_line_border(
    this_line: AhbLine, next_line: AhbLine, next_filled_segment: Optional[str], opening_segment_code: str
) -> bool:
    """
    returns true iff this_line and next_line are the border between two neighbouring segment groups with the same.
    This means that this_line is the last one in one segment group SGx and the next_line is in the first segment group
    SGx (note that both SGx are the same!)
    """
    # There are unit tests just for this method.
    # If you suspect the "group_by_segment_group" method to have a bug, check the tests for this method first.
    if this_line.section_name == next_line.section_name:
        # The first section of the next segment group (hopefully) never has the same section name as the last entry in
        # the previous group
        return False
    this_line_is_opening_segment = this_line.segment_code == opening_segment_code
    next_line_is_opening_segment = next_line == opening_segment_code
    this_line_has_empty_segment = this_line.segment_code is None
    if (
        (this_line_is_opening_segment is False)
        and (next_line_is_opening_segment is False)
        and (this_line_has_empty_segment is True)
    ):
        return False
    next_line_has_empty_segment = next_line.segment_code is None
    if next_line_is_opening_segment is False and next_line_has_empty_segment is False:
        return False
    return next_filled_segment == opening_segment_code


# pylint:disable=too-few-public-methods
@attrs.define(auto_attribs=True, kw_only=True, frozen=True)
class AhbLocationLayer:
    """
    The AhbLocation consists of multiple layers of information about the nesting of a line in the SegmentGroupHierarchy.
    one _AhbLocationLayer models exactly one of these nesting layers. It is used only internally in the AhbLocation.

    The AhbLocationLayer relates to the AhbLocation just as the EdifactStackLevel relates to the EdifactStack.
    """

    segment_group_key: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.matches_re(r"^SG\d+$"))
    )
    """
    The key of the segment group in which the line is located; use None for the root.
    Example: "SG4"
    """
    opening_segment_code: str = attrs.field(validator=attrs.validators.matches_re(r"^[A-Z]{3}$"))
    """
    The opening_segment_code is the first segment code of the respective SegmentGroupHierarchy.
    Example: "IDE"
    """
    opening_qualifier: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.matches_re(r"^[A-Z]*\d*$"))
    )
    """"
    The opening_qualifier is the qualifier used in the opening_segment of this segment group.
    It is used to distinguish this segment group from other segment groups with the same key.
    (e.g. to distinguish between different SG8 for different purposes but the same internal structure).
    Example: "24"
    """


# pylint:disable=too-few-public-methods
@attrs.define(kw_only=True, auto_attribs=True, frozen=True)
class AhbLocation:
    """
    An ahb location describes where in a SegmentGroupHierarchy an AhbLine is located.
    The position is determined by the opening segments and qualifiers that have "passed by" while iterating over the
    lines inside an AHB from top to bottom.
    The location is initialized when you start iterating over the AHB by providing the relevant SegmentGroupHierarchy.
    Example: Consider the line: "SG10, CAV, 7111, Wertegranularität" from FV2210 UTILMD AHBs.
    It is located in
    * root (None, UNH, None)
    * "Vorgang" (SG4, IDE, 24)
    * "OBIS-Daten der Marktlokation" (SG8, SEQ, Z02)
    * "Wertegranularität" (SG10, CCI, ZE4)

    The AhbLocation relates to the AHB just as the EdifactStack relates to the MIG.
    """

    layers: List[AhbLocationLayer] = attrs.field(
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(AhbLocationLayer),
            iterable_validator=attrs.validators.min_len(1),
        )
    )
    """
    the single layers that define the segment groups involved in the location
    """
    segment_code: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.matches_re(r"^[A-Z]{3}$")), default=None
    )
    """
    The segment code of an element. None for groups only.
    """

    data_element_id: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.matches_re(r"^\d{4}$")), default=None
    )
    """
    the ID of th data element (inside the segment group described by layers), e.g. "0062"
    """
    qualifier: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)), default=None
    )
    """
    qualifier to identify the matching data element if the data_element id is not unique
    """

    def is_sub_location_of(self, other: "AhbLocation") -> bool:
        """
        Returns true iff this (self) location is a sub position of the other provided location.
        ([Foo][0][Bar]).is_sub_position_of([Foo][0]) is true.
        """
        # this is kind of copied and pasted from EdifactStack.is_sub_stack_of
        # but any generalisation is probably not worth the loss of readability
        if len(other.layers) > len(self.layers):
            # self cannot be a sub path of other if other is "deeper"
            return False
        for layer_self, layer_other in zip(other.layers, self.layers):  # , strict=False):
            # strict is False because it's ok if we stop the iteration if self.levels is "exhausted"; explicit is better
            # the type-ignore for the strict=False is necessary for Python<3.10
            if layer_self != layer_other:
                return False
        # the iteration stopped meaning that for all layers that both other and self share, they are identical.
        # That's the definition of a sub location. It also means that any stack is a sub location of itself.
        return True

    def is_parent_of(self, other: "AhbLocation") -> bool:
        """
        Returns true iff this other stack is a sub position of self.
        """
        return other.is_sub_location_of(self)


@attrs.define(kw_only=True, frozen=True, auto_attribs=True)
class _PseudoAhbLocation(AhbLocation):
    """
    a separate class to distinguish _real_ ahb locations from fake/pseudo ahb locations.
    The latter should never be used outside this module.
    """


@overload
def find_common_ancestor(location_x: _PseudoAhbLocation, location_y: _PseudoAhbLocation) -> _PseudoAhbLocation:
    ...


@overload
def find_common_ancestor(
    location_x: _PseudoAhbLocation, location_y: Union[AhbLocation, _PseudoAhbLocation]
) -> _PseudoAhbLocation:
    ...


@overload
def find_common_ancestor(
    location_x: Union[AhbLocation, _PseudoAhbLocation], location_y: _PseudoAhbLocation
) -> _PseudoAhbLocation:
    ...


@overload
def find_common_ancestor(location_x: AhbLocation, location_y: AhbLocation) -> AhbLocation:
    ...


def find_common_ancestor(
    location_x: Union[AhbLocation, _PseudoAhbLocation], location_y: Union[AhbLocation, _PseudoAhbLocation]
) -> Union[AhbLocation, _PseudoAhbLocation]:
    """
    Finds the last common ancestor of location_x and location_y.
    If the layers of location X are:  [A,B,C,F,G,H]
    And the layers of location Y are: [A,B,C,D,E,F]
    then the last common ancestor is: [A,B,C]
    :param location_x:
    :param location_y:
    :return: the location that is the last common ancestor of x and y
    """
    result_layers: List[AhbLocationLayer] = []
    only_consider_sg_key = isinstance(location_x, _PseudoAhbLocation) or isinstance(location_y, _PseudoAhbLocation)
    for layer_x, layer_y in zip(location_x.layers, location_y.layers):
        if layer_x == layer_y or (
            only_consider_sg_key is True and layer_x.segment_group_key == layer_y.segment_group_key
        ):
            result_layers.append(layer_x)
        else:
            break
    try:
        if only_consider_sg_key:
            return _PseudoAhbLocation(layers=result_layers)
        return AhbLocation(layers=result_layers)
    except ValueError as value_error:
        # We raise an exception if len(result_layers) is 0 because this case shouldn't happen in locations that
        # originate from the same AHB.
        raise ValueError("There is no common ancestor") from value_error


def _construct_pseudo_location(
    sg_key: Optional[str],
    segment_group_hierarchy: SegmentGroupHierarchy,
    existing_layers: Optional[List[AhbLocationLayer]] = None,
) -> Optional[_PseudoAhbLocation]:
    """
    use the segment group hierarchy to _guess_ where the segment group with key sg_key is located
    :param sg_key:
    :param segment_group_hierarchy:
    :return:
    """
    this_layer = AhbLocationLayer(
        segment_group_key=segment_group_hierarchy.segment_group,
        opening_segment_code=segment_group_hierarchy.opening_segment,
        opening_qualifier=None,
    )
    layers: List[AhbLocationLayer]
    if existing_layers is None:
        layers = [this_layer]
    else:
        layers = existing_layers.copy()
        layers.append(this_layer)
    if sg_key == segment_group_hierarchy.segment_group:
        return _PseudoAhbLocation(layers=layers)
    for sub_hierarchy in segment_group_hierarchy.sub_hierarchy or []:
        sub_result = _construct_pseudo_location(sg_key, sub_hierarchy, layers)
        if sub_result is not None:
            return sub_result
    return None


def _find_common_ancestor_from_sgh(
    sg_key_x: Optional[str], sg_key_y: Optional[str], segment_group_hierarchy: SegmentGroupHierarchy
) -> _PseudoAhbLocation:
    """
    Finds the last common ancestor of the segment groups x and y.
    The function uses the provided segment group hierarchy to construct pseudo AHBLocations.
    These locations are then fed into the other overload of find_common_ancestor
    :param segment_group_hierarchy: the SGH used to locate the SGs
    :param sg_key_x: key of the segment group at location x
    :param sg_key_y: key of the segment group at location y
    :return: A Pseudo-Location that is a placeholder for the last common ancestor of x and y
    """
    location_x: Optional[_PseudoAhbLocation] = _construct_pseudo_location(sg_key_x, segment_group_hierarchy)
    if location_x is None:
        raise ValueError(f"Couldn't locate {sg_key_x} in the given hierarchy")
    location_y: Optional[_PseudoAhbLocation] = _construct_pseudo_location(sg_key_y, segment_group_hierarchy)
    if location_x is None:
        raise ValueError(f"Couldn't locate {sg_key_y} in the given hierarchy")
    # note that the result is generally complete. just the number of layers has a meaning
    return find_common_ancestor(location_x, location_y)  # type:ignore[arg-type]


@attrs.define(auto_attribs=True, frozen=True, kw_only=True)
class _AhbLocationDistance:
    """
    Describes the differences in the hierarchy between two locations.
    Assume the start location is layers=[A,B,C,D,E,F] and the target location is layers=[A,B,G,H].
    Then the distance is (with B as common ancestor):
    layers_up = 4 (F--1-->E--2-->D--3-->C--4-->B)
    layers_down = 2 (B--1-->G--2-->H)
    A distance with layers_down == layers_up == 1 means the two locations are sibling.
    A distance with layers_down == 0 means that the target location is a parent of the source location.
    A distance with layers_up == 0 means that the source location is a parent of the target location.
    """

    layers_up: int = attrs.field(
        validator=attrs.validators.and_(attrs.validators.instance_of(int), attrs.validators.ge(0))
    )
    """
    Describes the number of layers (up) between the source location and the common ancestor of both locations.
    Is always >= 0.
    """
    layers_down: int = attrs.field(
        validator=attrs.validators.and_(attrs.validators.instance_of(int), attrs.validators.ge(0))
    )
    """
    Describes the number of layer (down) between the common ancestor of both locations and the target location.
    Is always >= 0.
    """


def calculate_distance(location_x: AhbLocation, location_y: AhbLocation) -> _AhbLocationDistance:
    """
    Calculate the "distance" between two locations.
    The distance is measured in how many layers you have to go up and how many down reach location_y from location_y.
    For more information see the docstring of `_AhbLocationDistance`
    :param location_x: the start location
    :param location_y: the target location
    :return: the number of layers between location_x and location_y via their common ancestor.
    """
    common_ancestor = find_common_ancestor(location_x, location_y)
    return _AhbLocationDistance(
        layers_up=len(location_x.layers) - len(common_ancestor.layers),
        layers_down=len(location_y.layers) - len(common_ancestor.layers),
    )


class _DifferentialAhbLineHierarchyChange(Enum):
    """
    Describes the different scenarios that can happen when moving from one AHBline to the next.
    """

    STAY = 1  #: Stay in the same ("selbe") segment group; e.g. UTILMD SG4 IDE+24 -> SG4 IDE+24
    MOVE_TO_NEIGHBOUR_SAME_KEY = 2
    """
    Stay on the same hierarchy level but move to a neighbouring group that has the same segment group key.
    E.g. UTILMD SG2 NAD+MS -> SG2 NAD+MR
    """
    MOVE_TO_NEIGHBOUR_DIFFERENT_KEY = 4
    """
    Stay on the same hierarchy level but move to a neighbouring group with a different segment group key.
    E.g. UTILMD SG5 LOC -> SG6 RFF
    """
    DIVE_INTO_SUB_GROUP = 8
    """
    Move into a nested subgroup with a different segment group key.
    E.g. UTILMD SG2->SG3
    """
    LEAVE_TO_PARENT = 16
    """
    Leave a segment group back to the parent hierarchy level. E.g. UTILMD SG3 -> SG2
    """
    LEAVE_TO_PARENT_AND_DIVE_INTO_SUB_GROUP = DIVE_INTO_SUB_GROUP + LEAVE_TO_PARENT
    """
    often when leaving one sub group we directly dive into the next one.
    """

    def is_segment_group_change(self) -> bool:
        """
        returns true iff this item is denotes any segment groups change
        """
        return self != _DifferentialAhbLineHierarchyChange.STAY


def _this_line_is_hierarchically_below_the_previous_sg_key(
    this_line: AhbLine, previous_segment_group_key, segment_group_hierarchy: SegmentGroupHierarchy
) -> bool:
    """
    returns true iff this_line is hierarchically below the previous segment group key; false otherwise
    :param this_line: e.g. SG3 CTA
    :param previous_segment_group_key: SG2
    :param segment_group_hierarchy: [UNH, [SG2 [SG3]], [SG2], [SG4[SG...]]]
    :return: true if SG3 hierarchically line is below SG2
    """
    if this_line.segment_group_key == previous_segment_group_key:
        return False
    if segment_group_hierarchy.segment_group == previous_segment_group_key:
        # anything is below root
        return True
    return segment_group_hierarchy.sg_is_hierarchically_below(this_line.segment_group_key, previous_segment_group_key)


# pylint:disable=too-many-arguments
def _determine_hierarchy_change(
    this_ahb_line: AhbLine,
    previous_ahb_line: Optional[AhbLine],
    this_next_segment: str,
    this_next_qualifier: str,
    last_opening_segment: Optional[str],
    last_opening_qualifier: Optional[str],
    is_ready_for_sg_change: bool,
    segment_group_hierarchy: SegmentGroupHierarchy,
) -> _DifferentialAhbLineHierarchyChange:
    """
    Determine which kind of transition happens between two AHB lines
    :return: see `_DifferentialAhbLineHierarchyChange`
    """
    previous_sg_key: Optional[str] = None
    if previous_ahb_line is not None:
        previous_sg_key = previous_ahb_line.segment_group_key
    if this_ahb_line.segment_group_key == previous_sg_key:  # No segment group change (key);
        if this_next_segment != last_opening_segment:
            return _DifferentialAhbLineHierarchyChange.STAY
        # next segment opens a new group with the same key
        if this_next_qualifier == last_opening_qualifier or (
            # as long as we're in UNH, we stay in UNH
            last_opening_segment is None
            and this_next_segment == "UNH"
        ):
            return _DifferentialAhbLineHierarchyChange.STAY
        if is_ready_for_sg_change:
            return _DifferentialAhbLineHierarchyChange.MOVE_TO_NEIGHBOUR_SAME_KEY
        return _DifferentialAhbLineHierarchyChange.STAY

    # An SG (key) change happened. Now we have to distinguish. Did we switch...
    # * ...to a sub group nested inside (add layers)
    # * ...back to a parent group (remove layers)
    # * ...even both: move to parent (remove layer) and then dive into another subgroup (add layer) at the same time
    if segment_group_hierarchy.sg_is_hierarchically_below(this_ahb_line.segment_group_key, previous_sg_key):
        return _DifferentialAhbLineHierarchyChange.DIVE_INTO_SUB_GROUP
    if segment_group_hierarchy.sg_is_hierarchically_below(previous_sg_key, this_ahb_line.segment_group_key):
        # result = _DifferentialAhbLineHierarchyChange.LEAVE_TO_PARENT
        # let's check if we're back in the previous parent sg or already in another sub sg
        # we'd detect that we move into the next subgroup already
        # todo: does the case leave to parent really happen? maybe for unh only?
        pass
    return _DifferentialAhbLineHierarchyChange.LEAVE_TO_PARENT_AND_DIVE_INTO_SUB_GROUP  # sibling A->B->C to A-B->D


def determine_hierarchy_changes(
    ahb_lines: List[AhbLine], segment_group_hierarchy: SegmentGroupHierarchy
) -> List[Tuple[AhbLine, _DifferentialAhbLineHierarchyChange]]:
    """
    Determine the differential hierarchy changes between neighbouring ahb lines.
    The returned tuple at index n is the line n itself + the change in hierarchy necessary to get from line n-1 to n.
    """
    result: List[Tuple[AhbLine, _DifferentialAhbLineHierarchyChange]] = []
    segment_code_was_none = True
    previous_line: Optional[AhbLine] = None
    last_opening_qualifier: str = "UNH"
    last_opening_segment: Optional[str] = None
    zip_kwargs = {}
    if sys.version_info.minor >= 10:  # we implicitly assume python 3 here
        zip_kwargs = {"strict": True}  # strict=True has been introduced in 3.10
    for this_ahb_line, this_next_segment, this_next_qualifier in list(
        zip(
            ahb_lines,
            [x[1] for x in _enhance_with_next_segment(ahb_lines)],
            [x[1] for x in _enhance_with_next_value_pool_entry(ahb_lines)],
            **zip_kwargs,
        )
    ):
        if this_ahb_line.segment_code is None:
            segment_code_was_none = True
        change = _determine_hierarchy_change(
            this_ahb_line=this_ahb_line,
            previous_ahb_line=previous_line,
            this_next_segment=this_next_segment,
            this_next_qualifier=this_next_qualifier,
            last_opening_qualifier=last_opening_qualifier,
            last_opening_segment=last_opening_segment,
            is_ready_for_sg_change=segment_code_was_none,
            segment_group_hierarchy=segment_group_hierarchy,
        )
        if change.is_segment_group_change():
            last_opening_qualifier = this_next_qualifier
            last_opening_segment = this_next_segment
        if this_ahb_line.segment_code is not None:
            segment_code_was_none = False
        result.append((this_ahb_line, change))
        previous_line = this_ahb_line
    return result


def determine_locations(
    segment_group_hierarchy: SegmentGroupHierarchy, ahb_lines: List[AhbLine]
) -> List[Tuple[AhbLine, AhbLocation]]:
    """
    If you provide _all_ lines of an AHB from top to bottom, this function will enrich the list of ahb_lines with the
    respective locations of the single AHBLines. These locations can then be used to find/match the lines with the MIG.
    :param segment_group_hierarchy: the general structure of the MIG
    :param ahb_lines: all lines of an AHB
    :return: the same lines that have been entered but together with their location which is derived from the segment
    group hierarchy.
    """
    starting_layer = AhbLocationLayer(
        segment_group_key=segment_group_hierarchy.segment_group,
        opening_segment_code=segment_group_hierarchy.opening_segment,
        opening_qualifier=None,
    )
    layers: List[AhbLocationLayer] = [starting_layer]
    # the layers are the internal representation of the position.
    # layers[0] is always the SGH root.
    # layers[-1] is the segment group that is also denoted at the beginning of the current AHBLine.

    # Basically we now iterate over the ahb_lines lines and try to remember which segment groups we entered on our
    # way and which one we also exited so that in the end, we know where the last item of ahb_lines is located.
    # To account reduce the complexity in this function, we first enhance the lines with additional information:
    # 1. what is the next segment code?
    # 2. what is the next qualifier?
    # Using these helper functions saves us from looking ahead in our iteration.
    # We only need to look at one item at a time
    result: List[Tuple[AhbLine, AhbLocation]] = []
    for (this_ahb_line, change), this_next_segment, this_next_qualifier in zip(
        determine_hierarchy_changes(ahb_lines, segment_group_hierarchy),
        [x[1] for x in _enhance_with_next_segment(ahb_lines)],
        [x[1] for x in _enhance_with_next_value_pool_entry(ahb_lines)],
    ):
        if last(layers).opening_qualifier is None and len(result) == 0:
            layers[-1] = AhbLocationLayer(
                segment_group_key=segment_group_hierarchy.segment_group,
                opening_segment_code=segment_group_hierarchy.opening_segment,
                opening_qualifier=this_next_qualifier,
            )
        if change == _DifferentialAhbLineHierarchyChange.STAY:
            # No segment group change; The layers represent the location already.
            result.append(
                (
                    this_ahb_line,
                    AhbLocation(
                        layers=layers.copy(),
                        data_element_id=this_ahb_line.data_element,
                        segment_code=this_next_segment,
                        qualifier=this_ahb_line.value_pool_entry,
                    ),
                )
            )
            continue
        if change == _DifferentialAhbLineHierarchyChange.MOVE_TO_NEIGHBOUR_SAME_KEY:
            layers.pop()
            layers.append(
                AhbLocationLayer(
                    segment_group_key=this_ahb_line.segment_group_key,
                    opening_segment_code=this_next_segment,
                    opening_qualifier=this_next_qualifier,
                )
            )
            result.append(
                (
                    this_ahb_line,
                    AhbLocation(
                        layers=layers.copy(),
                        data_element_id=this_ahb_line.data_element,
                        segment_code=this_next_segment,
                        qualifier=this_ahb_line.value_pool_entry,
                    ),
                )
            )
            continue
        # switch to a sublevel or to a neighbouring group with a different SG key (SGx->SGy)
        if change == _DifferentialAhbLineHierarchyChange.DIVE_INTO_SUB_GROUP:
            # we moved into a subgroup, e.g. from UTILMD SG4 to SG5
            layers.append(
                AhbLocationLayer(
                    segment_group_key=this_ahb_line.segment_group_key,
                    opening_segment_code=this_next_segment,
                    opening_qualifier=this_next_qualifier,
                )
            )
            result.append(
                (
                    this_ahb_line,
                    AhbLocation(
                        layers=layers.copy(),
                        data_element_id=this_ahb_line.data_element,
                        segment_code=this_next_segment,
                        qualifier=this_ahb_line.value_pool_entry,
                    ),
                )
            )
            continue
        if change == _DifferentialAhbLineHierarchyChange.LEAVE_TO_PARENT_AND_DIVE_INTO_SUB_GROUP:
            # [SG2 (NAD+MS), SG3]->[SG2 (NAD+MR)] # one group up (leave SG3, then remove old SG2)
            # [SG4, SG8, SG10]-> [SG4,SG12] # two groups up! (leave SG10, leave SG8, then enter sg12)
            common_ancestor = _find_common_ancestor_from_sgh(
                last(last(result)[1].layers).segment_group_key, this_ahb_line.segment_group_key, segment_group_hierarchy
            )
            distance_to_common_ancestor = calculate_distance(last(result)[1], common_ancestor)
            for _ in range(0, distance_to_common_ancestor.layers_up):
                # actually: this should be a separate case in the differential change enum
                layers.pop()
                if last(layers).segment_group_key == this_ahb_line.segment_group_key:  # this removes the old SG2
                    layers.pop()
            layers.append(  # this adds the next SG2
                AhbLocationLayer(
                    segment_group_key=this_ahb_line.segment_group_key,
                    opening_segment_code=this_next_segment,
                    opening_qualifier=this_next_qualifier,
                )
            )
            result.append(
                (
                    this_ahb_line,
                    AhbLocation(
                        layers=layers.copy(),
                        data_element_id=this_ahb_line.data_element,
                        segment_code=this_next_segment,
                        qualifier=this_ahb_line.value_pool_entry,
                    ),
                )
            )
            continue
        if change == _DifferentialAhbLineHierarchyChange.LEAVE_TO_PARENT:
            raise NotImplementedError("todo think if this really happens")
    return result



"""
division helper is a module that allow to distinguish electricity ("Strom") and gas ("Gas")
"""
from enum import Enum
from typing import Optional


class Division(str, Enum):
    """
    the possibility to distinguish between electricity and gas
    """

    ELECTRICITY = "ELECTRICITY"  #: strom
    GAS = "GAS"  #: gas


def get_division(field_name: Optional[str]) -> Optional[Division]:
    """
    returns a division if the field name is definitly from that division, None otherwise
    """
    if not field_name:
        return None
    if is_gas_name(field_name) is True:  # type:ignore[union-attr]
        return Division.GAS
    if is_electricity_name(field_name) is True:  # type:ignore[union-attr]
        return Division.ELECTRICITY
    return None


def is_gas_name(field_name: str, _is_recursive: bool = False) -> Optional[bool]:
    """
    returns true iff the field name is definitely a gas field
    """
    if "gas" in field_name.lower():  # maybe use regex `\bgas\b` just to be sure
        return True
    if _is_recursive is False and is_electricity_name(field_name, _is_recursive=True) is True:
        return False
    return None


def is_electricity_name(field_name: str, _is_recursive: bool = False) -> Optional[bool]:
    """
    returns true iff the field name is definitely a strom field
    """
    if "strom" in field_name.lower():  # maybe use regex `\bstrom\b` just to be sure
        return True
    if _is_recursive is False and is_gas_name(field_name, _is_recursive=True) is True:
        return False
    return None



"""
MAUS is the MIG AHB Utility stack.
This module contains methods to merge data from Message Implementation Guide and Anwendungshandbuch
"""
from itertools import groupby
from typing import List, Optional, Sequence, Set, Tuple

from more_itertools import first, first_true, last

from maus.models.anwendungshandbuch import _VERSION, AhbLine, DeepAnwendungshandbuch, FlatAnwendungshandbuch
from maus.models.edifact_components import (
    DataElement,
    DataElementFreeText,
    DataElementValuePool,
    EdifactStack,
    Segment,
    SegmentGroup,
    ValuePoolEntry,
    derive_data_type_from_segment_code,
)
from maus.models.message_implementation_guide import SegmentGroupHierarchy
from maus.navigation import AhbLocation, calculate_distance, determine_locations
from maus.reader.mig_reader import MigReader


def merge_lines_with_same_data_element(
    ahb_lines: Sequence[AhbLine], first_stack: Optional[EdifactStack]
) -> DataElement:
    """
    Merges lines that have the same data element into a single data element instance which is returned
    """
    distinct_data_element_keys = {ahb_line.data_element for ahb_line in ahb_lines}
    if len(distinct_data_element_keys) != 1:
        raise ValueError(
            "You must only use this function with lines that share the same data element but the "
            f"parameter ahb_lines contains: {', '.join([x or '' for x in distinct_data_element_keys])} "
        )
    result: DataElement
    discriminator = None
    if first_stack is not None:
        discriminator = first_stack.to_json_path()
    if ahb_lines[0].value_pool_entry is not None:
        result = DataElementValuePool(
            discriminator=discriminator,
            value_pool=[],
            data_element_id=ahb_lines[0].data_element,  # type:ignore[arg-type]
            entered_input=None,
        )
        for data_element_value_entry in ahb_lines:
            if not data_element_value_entry.ahb_expression:
                # value pool entries with empty/None AHB expression shall not be included
                # https://github.com/Hochfrequenz/mig_ahb_utility_stack/issues/38
                continue
            value_pool_entry = ValuePoolEntry(
                qualifier=data_element_value_entry.value_pool_entry,  # type:ignore[arg-type]
                meaning=data_element_value_entry.name.strip(),  # type:ignore[assignment,union-attr]
                ahb_expression=data_element_value_entry.ahb_expression,
            )
            result.value_pool.append(value_pool_entry)  # type:ignore[index]
    else:
        result = DataElementFreeText(
            entered_input=None,
            ahb_expression=ahb_lines[0].ahb_expression,  # type:ignore[arg-type]
            discriminator=discriminator,
            data_element_id=first(  # type:ignore[union-attr]
                ahb_lines, lambda line: line.data_element is not None
            ).data_element,  # type:ignore[arg-type]
        )
        # a free text field never spans more than 1 line

        data_type = derive_data_type_from_segment_code(ahb_lines[0].segment_code)  # type:ignore[arg-type]
        if data_type is not None:
            result.value_type = data_type
    return result


def _remove_qualifier(location: AhbLocation) -> AhbLocation:
    """
    returns a copy of location but with an empty qualifier
    """
    result: AhbLocation = AhbLocation(
        qualifier=None,
        layers=location.layers,
        data_element_id=location.data_element_id,
        segment_code=location.segment_code,
    )
    return result


# I'm aware the function is too long; Let's first make it work, then split up into separate functions.
# pylint:disable=too-many-locals, too-many-branches, too-many-statements
# https://github.com/Hochfrequenz/mig_ahb_utility_stack/issues/205
def to_deep_ahb(
    flat_ahb: FlatAnwendungshandbuch, segment_group_hierarchy: SegmentGroupHierarchy, mig_reader: MigReader
) -> DeepAnwendungshandbuch:
    """
    Converts a flat ahb into a nested ahb using the provided segment hierarchy
    """
    result = DeepAnwendungshandbuch(meta=flat_ahb.meta, lines=[])
    result.meta.maus_version = _VERSION
    parent_group_lists: List[List[SegmentGroup]] = []
    used_stacks: Set[str] = set()
    # The following lists are _not_ a view into the lines that are going to follow (hence no name starting with "next")
    # Instead the lists are a reference to the respective attribute _inside_ the upper hierarchy element.
    # So whenever we e.g. run into a data element, we can "remember" to where the data element shall be appended.
    # The lists are our "reference" so that each element deeper into the hierarchy (e.g. data elements) knows to which
    # parent (e.g. segment) it belongs.
    # Note that the lists are never assigned to anything.
    # + they're always instantiated empty
    # + then we store a reference to the list in `append_next_foo_here`
    # + when ever we find a foo downstream we append it to this list (and thereby assign it to its parent)
    # + when ever we find a new parent element, this variable that holds the reference to the old (now previous) parent
    # is replaced with a reference to the new parent to all the downstream elements to come.
    append_next_segments_here: List[Segment]  #: is instantiated/replaced whenever a new segment group is created
    append_next_sg_here: List[SegmentGroup]  #: is instantiated/replaced whenever a new segment group is created
    append_next_data_elements_here: List[DataElement]  #: is instantiated/replaced whenever a new segment is created
    previous_position: AhbLocation
    for position, layer_group in groupby(
        determine_locations(segment_group_hierarchy, flat_ahb.lines),
        key=lambda line_and_position: _remove_qualifier(line_and_position[1]),
    ):
        layer_group: List[Tuple[AhbLine, AhbLocation]] = list(layer_group)  # type:ignore[no-redef]
        data_element_lines = [x[0] for x in layer_group]  # index 1 is the position
        if not any((True for line in data_element_lines if line.segment_code is not None)):
            continue  # section heading only
        stack: EdifactStack
        try:
            if len(data_element_lines) == 1:
                position = layer_group[0][1]  # type:ignore[index]
            stack = mig_reader.get_edifact_stack(position)
        except ValueError:
            # if the AHB/MIG matching does not work as expected, set your breakpoints here
            stack = None  # type:ignore[assignment]
            if len(data_element_lines) > 1:
                for _position in (x[1] for x in layer_group):
                    try:
                        stack = mig_reader.get_edifact_stack(layer_group[0][1])  # type:ignore[index]
                        break
                    except ValueError:
                        pass
        if any((True for line in data_element_lines if line.data_element is not None)):
            if not any((True for line in data_element_lines if line.ahb_expression is not None)):
                # if none of the items is marked with an ahb expression it's probably not required in this AHB
                continue
            data_element = merge_lines_with_same_data_element(data_element_lines, first_stack=stack)
            try:
                append_next_data_elements_here.append(data_element)  # pylint:disable=used-before-assignment
            except UnboundLocalError as unbound_local_error:
                raise ValueError(f"No segment has been created for {stack}") from unbound_local_error
        else:
            first_line = first(data_element_lines)
            last_line = last(data_element_lines)
            if (
                first_line.segment_group_key == last(position.layers).segment_group_key
                and last(position.layers).opening_segment_code == last_line.segment_code
                and stack is not None
                and stack.to_json_path() not in used_stacks
                and first_line.ahb_expression is not None
            ):
                # a new segment group has been opened
                segment_group = SegmentGroup(
                    discriminator=stack.to_json_path(),
                    # type:ignore[arg-type] # might be None now, will be replaced later
                    ahb_expression=first_line.ahb_expression.strip(),
                    segments=[],
                    segment_groups=[],
                    ahb_line_index=first_line.index,
                )
                used_stacks.add(stack.to_json_path())
                append_next_segments_here = segment_group.segments  # type:ignore[assignment]
                if segment_group.discriminator == '$["Dokument"][0]["Nachricht"][0]':
                    result.lines.append(segment_group)
                    append_next_sg_here = segment_group.segment_groups  # type:ignore[assignment]
                    parent_group_lists.append(result.lines)
                elif position.is_sub_location_of(previous_position):  # pylint:disable=used-before-assignment
                    append_next_sg_here.append(segment_group)
                    parent_group_lists.append(append_next_sg_here)
                    append_next_sg_here = segment_group.segment_groups  # type:ignore[assignment]
                else:
                    distance = calculate_distance(previous_position, position)
                    for _ in range(0, distance.layers_up):
                        try:
                            append_next_sg_here = parent_group_lists.pop()
                        except IndexError as index_error:
                            raise ValueError(f"Couldn't move from {previous_position} to {position}") from index_error
                    for _ in range(0, distance.layers_down - 1):
                        parent_group_lists.append(append_next_sg_here)
                        append_next_sg_here = last(append_next_sg_here).segment_groups  # type:ignore[assignment]
                    append_next_sg_here.append(segment_group)
                    parent_group_lists.append(append_next_sg_here)
                    append_next_sg_here = segment_group.segment_groups  # type:ignore[assignment]

            assert last_line.data_element is None
            assert last_line.segment_code is not None
            # these assertion are because we assume that the lines always come like this:
            # Section Heading
            # SGx Foo      <-- a line with only the segment code but no actual content; this is where we're right now
            first_expression_line: Optional[AhbLine] = first_true(
                data_element_lines, default=None, pred=lambda l: l is not None and l.ahb_expression
            )
            if first_expression_line is None:
                continue
            first_expression: str = first_expression_line.ahb_expression  # type:ignore[assignment]
            discriminator: str
            if stack is None:
                # this is a pretty ugly hack. when we cannot find a position in the MIG, then we just stringify the
                # position and by doing so hope, that the reader of the maus can use it somehow
                discriminator = str(position)
            else:
                # this should be the default path
                discriminator = stack.to_json_path()
            segment = Segment(
                discriminator=discriminator,  # todo: sometimes the discriminator is not as sharp as it could have been
                data_elements=[],
                ahb_expression=first_expression,
                section_name=first_line.section_name,
                ahb_line_index=first_line.index,
            )
            append_next_data_elements_here = segment.data_elements
            append_next_segments_here.append(segment)
        previous_position = position
    return result
